/*
 *  AME_Queue.c
 *  Acxxx
 *
 *  Copyright (c) 2004-2014 Accusys, Inc. All rights reserved.
 *
 */


#include "AME_module.h"
#include "AME_import.h"
#include "AME_Queue.h"

#if defined (AME_MAC)
    #include "Acxxx.h"
#endif


void AME_Queue_Init_List(pAME_Queue_List pHead)
{
   pHead->pNext = pHead;
   pHead->pPrev = pHead;
   
   return;
}

void AME_Queue_add_head(pAME_Queue_List pHead,pAME_Queue_List pNew)
{
    pAME_Queue_List pNext = (pAME_Queue_List)pHead->pNext;
    
    pNext->pPrev = pNew;
    pNew->pNext = pNext;
    pNew->pPrev = pHead;
    pHead->pNext = pNew;
    
    return;
}


void AME_Queue_add_tail(pAME_Queue_List pHead,pAME_Queue_List pNew)
{
    pAME_Queue_List pPrev = (pAME_Queue_List)pHead->pPrev;
    
    pHead->pPrev = pNew;
    pNew->pNext = pHead;
    pNew->pPrev = pPrev;
    pPrev->pNext = pNew;
    
    return;
}

void AME_Queue_del_entry(pAME_Queue_List pEntry)
{
    pAME_Queue_List pPrev = (pAME_Queue_List)pEntry->pPrev;
    pAME_Queue_List pNext = (pAME_Queue_List)pEntry->pNext;
    
    pNext->pPrev = pPrev;
    pPrev->pNext = pNext;
    
    return;
}


AME_U32 AME_Queue_empty(pAME_Queue_List pHead)
{
    if (pHead->pNext == pHead) {
        return TRUE;
    } else {
        return FALSE;
    }
    
}

AME_U32 AME_Queue_Get_Max_cmd(AME_U32 ReqType)
{
    AME_U32 MAX_concurrent_command = 0;
    
    switch (ReqType)
    {
        case ReqType_DAS_SCSI:
            MAX_concurrent_command = MAX_NT_2208_concurrent_command;
            break;
        case ReqType_NT_SCSI_341:
            MAX_concurrent_command = MAX_NT_341_concurrent_command;
            break;
        case ReqType_NT_SCSI_2208:
            MAX_concurrent_command = MAX_NT_2208_concurrent_command;
            break;
        case ReqType_NT_SCSI_3108:
            MAX_concurrent_command = MAX_NT_3108_concurrent_command;
            break;
        case ReqType_NT_SCSI_3108_V2:
            MAX_concurrent_command = MAX_NT_3108_concurrent_command_V2;
            break;
        case ReqType_NT_SCSI_3108_R50:
            MAX_concurrent_command = MAX_NT_3108_concurrent_command_R50;
            break;
        default:
            break;
    }
    
    return MAX_concurrent_command;
}


AME_U32 AME_Queue_Data_Init(pAME_Data pAMEData,pAME_Queue_DATA pAMEQueueData)
{
    AME_U32 i;
    
    pAMEQueueData->pAMEData = pAMEData;
    pAMEData->pAMEQueueData = pAMEQueueData;
    
    AME_Queue_Init_List(&pAMEQueueData->AME_Queue_List_Free);
    AME_Queue_Init_List(&pAMEQueueData->AME_Queue_List_SCSI_Wait);
    AME_Queue_Init_List(&pAMEQueueData->AME_Queue_List_SCSI_Sending);
    AME_Queue_Init_List(&pAMEQueueData->AME_Queue_List_InBand);
    
    pAMEQueueData->Current_InBand_ReqType = 0;
    pAMEQueueData->Current_InBand_Raid_ID = 0;
    pAMEQueueData->Current_InBand_Is_Abort = FALSE;
    pAMEQueueData->Current_InBand_Request_Buffer_Phy_addr = 0;
    pAMEQueueData->Current_InBand_InBand_callback = NULL;
    pAMEQueueData->AME_Queue_InBand_Count = 0;
    
    for (i=0; i<MAX_RAID_System_ID; i++) {
        pAMEQueueData->AME_Queue_Raid_CMD_Count[i] = 0;
    }
    
    for (i=0; i<AME_Queue_Support_Cmd; i++)
    {
        pAMEQueueData->AMEQueueBuffer[i].ReqInUse = FALSE;
        pAMEQueueData->AMEQueueBuffer[i].Raid_ID  = 0;
        pAMEQueueData->AMEQueueBuffer[i].Buffer_ID  = i;
        pAMEQueueData->AMEQueueBuffer[i].ReqType  = 0;
        pAMEQueueData->AMEQueueBuffer[i].Request_Buffer_Phy_addr = 0;
        pAMEQueueData->AMEQueueBuffer[i].callback = NULL;
        
        AME_Queue_add_tail(&pAMEQueueData->AME_Queue_List_Free,&pAMEQueueData->AMEQueueBuffer[i].AMEQueue_list);
    }
    
    return TRUE;
}

AME_U32 AME_Queue_Data_ReInit(pAME_Data pAMEData)
{
    AME_U32 i;
    pAME_Queue_DATA pAMEQueueData = (pAME_Queue_DATA)pAMEData->pAMEQueueData;
    
    AME_spinlock(pAMEData);
    
    AME_Queue_Init_List(&pAMEQueueData->AME_Queue_List_Free);
    AME_Queue_Init_List(&pAMEQueueData->AME_Queue_List_SCSI_Wait);
    AME_Queue_Init_List(&pAMEQueueData->AME_Queue_List_SCSI_Sending);
    AME_Queue_Init_List(&pAMEQueueData->AME_Queue_List_InBand);
    
    pAMEQueueData->Current_InBand_ReqType = 0;
    pAMEQueueData->Current_InBand_Raid_ID = 0;
    pAMEQueueData->Current_InBand_Is_Abort = FALSE;
    pAMEQueueData->Current_InBand_Request_Buffer_Phy_addr = 0;
    pAMEQueueData->Current_InBand_InBand_callback = NULL;
    pAMEQueueData->AME_Queue_InBand_Count = 0;
    
    for (i=0; i<MAX_RAID_System_ID; i++) {
        pAMEQueueData->AME_Queue_Raid_CMD_Count[i] = 0;
    }
    
    for (i=0; i<AME_Queue_Support_Cmd; i++)
    {
        pAMEQueueData->AMEQueueBuffer[i].ReqInUse = FALSE;
        pAMEQueueData->AMEQueueBuffer[i].Raid_ID  = 0;
        pAMEQueueData->AMEQueueBuffer[i].Buffer_ID  = i;
        pAMEQueueData->AMEQueueBuffer[i].ReqType  = 0;
        pAMEQueueData->AMEQueueBuffer[i].Request_Buffer_Phy_addr = 0;
        pAMEQueueData->AMEQueueBuffer[i].callback = NULL;
        
        AME_Queue_add_tail(&pAMEQueueData->AME_Queue_List_Free,&pAMEQueueData->AMEQueueBuffer[i].AMEQueue_list);
    }
    
    AME_spinunlock(pAMEData);
    
    return TRUE;
}

AME_U32 AME_Queue_Clean_Error_Raid_CMD_Count(pAME_Data pAMEData,AME_U32 RAID_ID)
{
    pAME_Queue_DATA pAMEQueueData = (pAME_Queue_DATA)pAMEData->pAMEQueueData;
    
    AME_spinlock(pAMEData);
    pAMEQueueData->AME_Queue_Raid_CMD_Count[RAID_ID] = 0;
    AME_spinunlock(pAMEData);
    
    return TRUE;
}


AME_U32 AME_Queue_Get_QCount(pAME_Queue_DATA pAMEQueueData)
{
    pAME_Queue_Buffer pAMEQueueBuffer;
    pAME_Data pAMEData = (pAME_Data)pAMEQueueData->pAMEData;
    pAME_Queue_List pHead = &pAMEQueueData->AME_Queue_List_SCSI_Sending;
    pAME_Queue_List pNext = pHead;   
    AME_U32 count = 0;
    AME_U64 Req_Phy_addr = 0;
    
    if (TRUE == AME_Queue_empty(pHead)) {
        return FALSE;
    }
    
    while (1)
    {       
        pNext = (pAME_Queue_List)pNext->pNext;
        
        if (pNext == pHead) {
            return count;
        }            
        
        count++;
        
        pAMEQueueBuffer = (pAME_Queue_Buffer)pNext;
        
        AME_Msg("!!!Evan:index[%d] , Raid_ID = %d , Request_Buffer_Phy_addr = 0x%x\n",count,pAMEQueueBuffer->Raid_ID,pAMEQueueBuffer->Request_Buffer_Phy_addr);
        
    }
    
}


AME_U32 AME_Queue_Get_Pended_SCSI_Command(pAME_Queue_DATA pAMEQueueData,pAME_Queue_Request pAMEQueueRequest,pAME_Queue_List *pNode)
{
    pAME_Queue_Buffer pAMEQueueBuffer;
    pAME_Queue_List pHead = &pAMEQueueData->AME_Queue_List_SCSI_Wait;
    pAME_Queue_List pNext  = (pAME_Queue_List)pHead->pNext;
    
    AME_U32 index = 0; 
    
    *pNode = pNext;
    
    if (TRUE == AME_Queue_empty(pHead)) {
        return FALSE;
    }
    
    AME_Queue_del_entry(pNext);
    
    pAMEQueueBuffer = (pAME_Queue_Buffer)pNext;
    pAMEQueueRequest->Raid_ID = pAMEQueueBuffer->Raid_ID;
    pAMEQueueRequest->ReqType = pAMEQueueBuffer->ReqType;
    pAMEQueueRequest->Request_Buffer_Phy_addr = pAMEQueueBuffer->Request_Buffer_Phy_addr;
    
    *pNode = pNext;
    
    //AME_Queue_add_tail(&pAMEQueueData->AME_Queue_List_SCSI_Sending,pNext);
    
    return TRUE;
}


AME_U32 AME_Queue_Get_Sended_SCSI_Command(pAME_Queue_DATA pAMEQueueData,AME_U64 Request_Buffer_Phy_addr,AME_U32 Raid_ID)
{
    pAME_Queue_Buffer pAMEQueueBuffer;
    pAME_Data pAMEData = (pAME_Data)pAMEQueueData->pAMEData;
    pAME_Queue_List pHead = &pAMEQueueData->AME_Queue_List_SCSI_Sending;
    pAME_Queue_List pNext = pHead;   
    AME_U32 index = 0;
    AME_U64 Req_Phy_addr = 0;
    
    if (TRUE == AME_Queue_empty(pHead)) {
        return FALSE;
    }
    
    while (1)
    {
        pNext = (pAME_Queue_List)pNext->pNext;
        
        if (pNext == pHead) {
            return FALSE;
        }
        
        index++;
        
        pAMEQueueBuffer = (pAME_Queue_Buffer)pNext;
        Req_Phy_addr = pAMEQueueBuffer->Request_Buffer_Phy_addr;
        
        // AME_Check_is_3108_new_type Request_Buffer_Phy_addr not address, used other check mathod(check Buffer ID).
        if (AME_Check_is_3108_new_type(pAMEData, Raid_ID) == TRUE)
        {
            // SgCount            :  8;     bits  0:7
            // PCISerialNumber    :  5;     bits  8:12
            // Reserved1          :  3;     bits  13:15
            // AddressIndex       : 16;     bits  16:31 ---> BufferID
            Req_Phy_addr = (Req_Phy_addr&0xFFFF0000) >> 16;
        }
        
        //AME_Msg("Sam Debug:  Req_Phy_addr 0x%x,  Request_Buffer_Phy_addr 0x%x\n",Req_Phy_addr,Request_Buffer_Phy_addr);
        if (Request_Buffer_Phy_addr == Req_Phy_addr) {
            AME_Queue_del_entry(pNext);
            
            // Clear QueueBuffer
            pAMEQueueBuffer->ReqInUse = FALSE;
            pAMEQueueBuffer->Raid_ID  = 0;
            pAMEQueueBuffer->ReqType  = 0;
            pAMEQueueBuffer->Request_Buffer_Phy_addr = 0;
            
            AME_Queue_add_tail(&pAMEQueueData->AME_Queue_List_Free,pNext);

            return TRUE;
        }
        
    }
    
}

AME_U32 AME_Queue_pending_SCSI_Command(pAME_Queue_DATA pAMEQueueData,AME_U32 RAID_ID,AME_U32 ReqType,AME_U64 Request_Buffer_Phy_addr)
{
    pAME_Queue_Buffer pAMEQueueBuffer;
    pAME_Queue_List pHead = &pAMEQueueData->AME_Queue_List_Free;
    pAME_Queue_List pNew  = (pAME_Queue_List)pHead->pNext;
    
    
    if (TRUE == AME_Queue_empty(pHead)) {
        AME_Msg_Err("AME Queue Error: AME_Queue_pending_SCSI_Command , AME_Queue_List_Free AME Queue Buffer empty!\n");            
        return FALSE;
    }
    
    AME_Queue_del_entry(pNew);
    
    pAMEQueueBuffer = (pAME_Queue_Buffer)pNew;
    pAMEQueueBuffer->ReqInUse = TRUE;
    pAMEQueueBuffer->Raid_ID  = RAID_ID;
    pAMEQueueBuffer->ReqType  = ReqType;
    pAMEQueueBuffer->Request_Buffer_Phy_addr = Request_Buffer_Phy_addr;
    
    AME_Queue_add_tail(&pAMEQueueData->AME_Queue_List_SCSI_Wait,pNew);   
    
    return TRUE;
}


AME_U32 AME_Queue_sending_SCSI_Command(pAME_Queue_DATA pAMEQueueData,AME_U32 RAID_ID,AME_U32 ReqType,AME_U64 Request_Buffer_Phy_addr)
{
    pAME_Queue_Buffer pAMEQueueBuffer;
    pAME_Queue_List pHead = &pAMEQueueData->AME_Queue_List_Free;
    pAME_Queue_List pNew  = (pAME_Queue_List)pHead->pNext;
    
    AME_U32 index = 0;
    pAME_Queue_List pNext;   
    
    if (TRUE == AME_Queue_empty(pHead)) {
        AME_Msg_Err("AME Queue Error: AME_Queue_sending_SCSI_Command , AME_Queue_List_Free AME Queue Buffer empty!\n");           
        return FALSE;
    }
    
    AME_Queue_del_entry(pNew);
    
    pAMEQueueBuffer = (pAME_Queue_Buffer)pNew;
    pAMEQueueBuffer->ReqInUse = TRUE;
    pAMEQueueBuffer->Raid_ID  = RAID_ID;
    pAMEQueueBuffer->ReqType  = ReqType;
    pAMEQueueBuffer->Request_Buffer_Phy_addr = Request_Buffer_Phy_addr;
    
    AME_Queue_add_tail(&pAMEQueueData->AME_Queue_List_SCSI_Sending,pNew);
    
    
    return TRUE;
}

AME_U32 AME_Queue_Repending_SCSI_Command(pAME_Queue_DATA pAMEQueueData,AME_U64 Request_Buffer_Phy_addr)
{
    pAME_Queue_Buffer pAMEQueueBuffer;
    pAME_Data pAMEData = (pAME_Data)pAMEQueueData->pAMEData;
    pAME_Queue_List pHead = &pAMEQueueData->AME_Queue_List_SCSI_Sending;
    pAME_Queue_List pNext = pHead;   
    AME_U32 index = 0;
    AME_U64 Req_Phy_addr = 0;
    
    if (TRUE == AME_Queue_empty(pHead)) {
        return FALSE;
    }
    
    while (1)
    {
        pNext = (pAME_Queue_List)pNext->pNext;
        
        if (pNext == pHead) {
            return FALSE;
        }
        
        index++;
        
        pAMEQueueBuffer = (pAME_Queue_Buffer)pNext;
        Req_Phy_addr = pAMEQueueBuffer->Request_Buffer_Phy_addr;
        
        //AME_Msg("Sam Debug:  Req_Phy_addr 0x%x,  Request_Buffer_Phy_addr 0x%x\n",Req_Phy_addr,Request_Buffer_Phy_addr);
        if (Req_Phy_addr == Request_Buffer_Phy_addr) {
            
            AME_Queue_del_entry(pNext);
            
            AME_Queue_add_tail(&pAMEQueueData->AME_Queue_List_SCSI_Wait,pNext);

            return TRUE;
        }
        
    }
}

AME_U32 AME_Queue_Resending_SCSI_Command(pAME_Data pAMEData,AME_U32 RAID_ID)
{
    pAME_Queue_DATA pAMEQueueData = (pAME_Queue_DATA)pAMEData->pAMEQueueData;
    pAME_Queue_Buffer pAMEQueueBuffer;
    pAME_Queue_List pHead = &pAMEQueueData->AME_Queue_List_SCSI_Sending;
    pAME_Queue_List pNext = pHead;  
    AME_U32 returnstatus = TRUE;  
   
    
    while (1)
    {
        pNext = (pAME_Queue_List)pNext->pNext;
        
        if (pNext == pHead) {
            break;
        }
        
        pAMEQueueBuffer = (pAME_Queue_Buffer)pNext;
        
        if(pAMEQueueBuffer->Raid_ID == RAID_ID) {
            if(TRUE == AME_Check_is_NT_Mode(pAMEData)) {
                returnstatus = AME_NT_write_Host_Inbound_Queue_Port_0x40(pAMEData,pAMEQueueBuffer->Raid_ID,pAMEQueueBuffer->Request_Buffer_Phy_addr);
            } else {
                returnstatus = AME_write_Host_Inbound_Queue_Port_0x40(pAMEData,(AME_U32)pAMEQueueBuffer->Request_Buffer_Phy_addr);
            }
            if(FALSE == returnstatus) {       
                AME_Queue_del_entry(pNext);
                AME_Queue_add_head(&pAMEQueueData->AME_Queue_List_SCSI_Wait,pNext);
                pNext = (pAME_Queue_List)pHead->pNext;
            } 
        }
      
    }    
    
    return returnstatus;
}


PVOID AME_Queue_Get_InBand_Command_callback(pAME_Data pAMEData)
{
    pAME_Queue_DATA pAMEQueueData = (pAME_Queue_DATA)pAMEData->pAMEQueueData;
    
    return pAMEQueueData->Current_InBand_InBand_callback;
    
}


AME_U32 AME_Queue_Get_Pended_InBand_Command(pAME_Queue_DATA pAMEQueueData,pAME_Queue_Request pAMEQueueRequest)
{
    pAME_Queue_Buffer pAMEQueueBuffer;
    pAME_Queue_List pHead = &pAMEQueueData->AME_Queue_List_InBand;
    pAME_Queue_List pNew  = (pAME_Queue_List)pHead->pNext;
    
    if (TRUE == AME_Queue_empty(pHead)) {
        return FALSE;
    }
    
    AME_Queue_del_entry(pNew);
    
    pAMEQueueBuffer = (pAME_Queue_Buffer)pNew;
    pAMEQueueRequest->Raid_ID = pAMEQueueBuffer->Raid_ID;
    pAMEQueueRequest->ReqType = pAMEQueueBuffer->ReqType;
    pAMEQueueRequest->Request_Buffer_Phy_addr = pAMEQueueBuffer->Request_Buffer_Phy_addr;
    pAMEQueueRequest->callback = pAMEQueueBuffer->callback;
    
    // Clear QueueBuffer
    pAMEQueueBuffer->ReqInUse = FALSE;
    pAMEQueueBuffer->Raid_ID  = 0;
    pAMEQueueBuffer->ReqType  = 0;
    pAMEQueueBuffer->Request_Buffer_Phy_addr = 0;
    pAMEQueueBuffer->callback = NULL;
    
    AME_Queue_add_tail(&pAMEQueueData->AME_Queue_List_Free,pNew);
    
    return TRUE;
}


AME_U32 AME_Queue_pending_InBand_Command(pAME_Queue_DATA pAMEQueueData,AME_U32 RAID_ID,AME_U32 ReqType,AME_U64 Request_Buffer_Phy_addr,PVOID callback)
{
    pAME_Queue_Buffer pAMEQueueBuffer;
    pAME_Queue_List pHead = &pAMEQueueData->AME_Queue_List_Free;
    pAME_Queue_List pNew  = (pAME_Queue_List)pHead->pNext;
    
    
    if (TRUE == AME_Queue_empty(pHead)) {
        AME_Msg_Err("AME Queue Error: AME_Queue_pending_InBand_Command , AME_Queue_List_Free AME Queue Buffer empty!\n");
        return FALSE;
    }
    
    AME_Queue_del_entry(pNew);
    
    pAMEQueueBuffer = (pAME_Queue_Buffer)pNew;
    pAMEQueueBuffer->ReqInUse = TRUE;
    pAMEQueueBuffer->Raid_ID  = RAID_ID;
    pAMEQueueBuffer->ReqType  = ReqType;
    pAMEQueueBuffer->Request_Buffer_Phy_addr = Request_Buffer_Phy_addr;
    pAMEQueueBuffer->callback  = callback;
    
    if (ReqType == ReqType_DAS_InBand_external) {
        AME_Queue_add_head(&pAMEQueueData->AME_Queue_List_InBand,pNew);
    } else {
        AME_Queue_add_tail(&pAMEQueueData->AME_Queue_List_InBand,pNew);
    }
    
    return TRUE;
    
}


AME_U32 AME_Queue_Repending_InBand_Command(pAME_Queue_DATA pAMEQueueData,AME_U32 RAID_ID,AME_U32 ReqType,AME_U64 Request_Buffer_Phy_addr,PVOID callback)
{
    pAME_Queue_Buffer pAMEQueueBuffer;
    pAME_Queue_List pHead = &pAMEQueueData->AME_Queue_List_Free;
    pAME_Queue_List pNew  = (pAME_Queue_List)pHead->pNext;
    
    if (TRUE == pAMEQueueData->Current_InBand_Is_Abort) {
        return TRUE;
    }
    
    
    if (TRUE == AME_Queue_empty(pHead)) {
        AME_Msg_Err("AME Queue Error: AME_Queue_Repending_InBand_Command , AME_Queue_List_Free AME Queue Buffer empty!\n");
        return FALSE;
    }
    
    AME_Queue_del_entry(pNew);
    
    pAMEQueueBuffer = (pAME_Queue_Buffer)pNew;
    pAMEQueueBuffer->ReqInUse = TRUE;
    pAMEQueueBuffer->Raid_ID  = pAMEQueueData->Current_InBand_Raid_ID;
    pAMEQueueBuffer->ReqType  = pAMEQueueData->Current_InBand_ReqType;
    pAMEQueueBuffer->Request_Buffer_Phy_addr = pAMEQueueData->Current_InBand_Request_Buffer_Phy_addr;
    pAMEQueueBuffer->callback  = pAMEQueueData->Current_InBand_InBand_callback;
    
    AME_Queue_add_head(&pAMEQueueData->AME_Queue_List_InBand,pNew);
    
    return TRUE;
    
}


/*
 * Function:  AME_Queue_Command
 * Description:  AME Queue Command support two command queue.
 *   1. SCSI command queue in NT environmet.
 *   2. InBand command queue in DAS environmet.
 */
AME_U32 AME_Queue_Command(pAME_Data pAMEData,AME_U32 RAID_ID,AME_U32 ReqType,AME_U64 Request_Buffer_Phy_addr,PVOID callback)
{
    AME_U32     returnstatus;
    AME_U32     MAX_concurrent_command = 0;
    pAME_Queue_DATA pAMEQueueData = (pAME_Queue_DATA)pAMEData->pAMEQueueData;
   
    
    switch (ReqType)
    {
        case ReqType_NT_InBand_internal:
        case ReqType_NT_InBand_external:
        case ReqType_DAS_InBand_internal:
        case ReqType_DAS_InBand_external:
            if (0 == pAMEQueueData->AME_Queue_InBand_Count)
            {
                pAMEQueueData->Current_InBand_ReqType = ReqType;
                pAMEQueueData->Current_InBand_Raid_ID = RAID_ID;
                pAMEQueueData->Current_InBand_Is_Abort = FALSE;
                pAMEQueueData->Current_InBand_Request_Buffer_Phy_addr = Request_Buffer_Phy_addr;
                pAMEQueueData->Current_InBand_InBand_callback = callback;                
                
                if ((ReqType == ReqType_NT_InBand_internal) || (ReqType == ReqType_NT_InBand_external)) {
                    returnstatus = AME_NT_write_Host_Inbound_Queue_Port_0x40(pAMEData,RAID_ID,Request_Buffer_Phy_addr);    
                } else {
                    returnstatus = AME_write_Host_Inbound_Queue_Port_0x40(pAMEData,(AME_U32)Request_Buffer_Phy_addr);
                }
                
                if (TRUE == returnstatus) {
                    AME_spinlock(pAMEData);
			        pAMEQueueData->AME_Queue_InBand_Count++;
			        AME_spinunlock(pAMEData);
			    }
			    
            }
            else {
                
                AME_spinlock(pAMEData);
                returnstatus = AME_Queue_pending_InBand_Command(pAMEQueueData,RAID_ID,ReqType,Request_Buffer_Phy_addr,callback);    
                AME_spinunlock(pAMEData);
                
            }
            break;
            
        case ReqType_Current_InBand_error:
            AME_Msg_Err("Current_InBand_Cmd error retry again\n");
            AME_spinlock(pAMEData);
            returnstatus = AME_Queue_Repending_InBand_Command(pAMEQueueData,RAID_ID,ReqType,Request_Buffer_Phy_addr,callback);
            AME_spinunlock(pAMEData);
            break;
        
        case ReqType_NT_SCSI_341:
        case ReqType_NT_SCSI_2208:
        case ReqType_NT_SCSI_3108:
        case ReqType_NT_SCSI_3108_V2:
            MAX_concurrent_command = AME_Queue_Get_Max_cmd(ReqType);
            if (MAX_concurrent_command == 0) {
                AME_Msg_Err("AME_Queue_Command:  Get unknow pended scsi cmd type (0x%x)\n",ReqType);
            }
            
            if ((pAMEQueueData->AME_Queue_Raid_CMD_Count[RAID_ID] < MAX_concurrent_command) && 
                (AME_NT_Raid_Bar_lost_Check(pAMEData,RAID_ID) == FALSE))
            {               
                AME_spinlock(pAMEData);
		        pAMEQueueData->AME_Queue_Raid_CMD_Count[RAID_ID]++;
		        returnstatus = AME_Queue_sending_SCSI_Command(pAMEQueueData,RAID_ID,ReqType,Request_Buffer_Phy_addr);
		        AME_spinunlock(pAMEData);
                
                returnstatus = AME_NT_write_Host_Inbound_Queue_Port_0x40(pAMEData,RAID_ID,Request_Buffer_Phy_addr);
                if (FALSE == returnstatus) {
			        AME_spinlock(pAMEData);
			        pAMEQueueData->AME_Queue_Raid_CMD_Count[RAID_ID]--;
                    if (TRUE == AME_Queue_Repending_SCSI_Command(pAMEQueueData,Request_Buffer_Phy_addr)) {
                        returnstatus = TRUE;
                    }
                    AME_spinunlock(pAMEData);
			    }	    
			
            } else {   
                AME_spinlock(pAMEData);
                returnstatus = AME_Queue_pending_SCSI_Command(pAMEQueueData,RAID_ID,ReqType,Request_Buffer_Phy_addr);
                AME_spinunlock(pAMEData); 
            }
            break;
            
        case ReqType_NT_SCSI_3108_R50:
            MAX_concurrent_command = AME_Queue_Get_Max_cmd(ReqType);
            if (MAX_concurrent_command == 0) {
                AME_Msg_Err("AME_Queue_Command:  Get unknow pended scsi cmd type (0x%x)\n",ReqType);
            }
            
            if ((pAMEQueueData->AME_Queue_Raid_CMD_Count[RAID_ID] < MAX_concurrent_command) && 
                (AME_NT_Raid_Bar_lost_Check(pAMEData,RAID_ID) == FALSE))
            {               
                AME_spinlock(pAMEData);
                pAMEQueueData->AME_Queue_Raid_CMD_Count[0]++; //R50 scsi cmd, count to Master Raid
		        returnstatus = AME_Queue_sending_SCSI_Command(pAMEQueueData,RAID_ID,ReqType,Request_Buffer_Phy_addr);
		        AME_spinunlock(pAMEData);
                
                returnstatus = AME_NT_write_Host_Inbound_Queue_Port_0x40(pAMEData,RAID_ID,Request_Buffer_Phy_addr);
                if (FALSE == returnstatus) {
			        AME_spinlock(pAMEData);
                    pAMEQueueData->AME_Queue_Raid_CMD_Count[0]--; //R50 scsi cmd, count to Master Raid
			        
                    if (TRUE == AME_Queue_Repending_SCSI_Command(pAMEQueueData,Request_Buffer_Phy_addr)) {
                        returnstatus = TRUE;
                    }
                    AME_spinunlock(pAMEData);
                    
			    }	    
			
            } else {   
                AME_spinlock(pAMEData);
                returnstatus = AME_Queue_pending_SCSI_Command(pAMEQueueData,RAID_ID,ReqType,Request_Buffer_Phy_addr);
                AME_spinunlock(pAMEData); 
            }
            break;
            
        case ReqType_DAS_SCSI:
            MAX_concurrent_command = AME_Queue_Get_Max_cmd(ReqType);
            if (MAX_concurrent_command == 0) {
                AME_Msg_Err("AME_Queue_Command:  Get unknow pended scsi cmd type (0x%x)\n",ReqType);
            }
            
            if ((pAMEQueueData->AME_Queue_Raid_CMD_Count[RAID_ID] < MAX_concurrent_command) && 
                (AME_Raid_2208_3108_Bar_lost_Check(pAMEData) == FALSE))
            {
                AME_spinlock(pAMEData);
		        pAMEQueueData->AME_Queue_Raid_CMD_Count[RAID_ID]++;
		        returnstatus = AME_Queue_sending_SCSI_Command(pAMEQueueData,RAID_ID,ReqType,Request_Buffer_Phy_addr);
		        AME_spinunlock(pAMEData);
                
                returnstatus = AME_write_Host_Inbound_Queue_Port_0x40(pAMEData,(AME_U32)Request_Buffer_Phy_addr);
    
			    if (FALSE == returnstatus) {
			        AME_spinlock(pAMEData);
			        pAMEQueueData->AME_Queue_Raid_CMD_Count[RAID_ID]--;
                    if (TRUE == AME_Queue_Repending_SCSI_Command(pAMEQueueData,Request_Buffer_Phy_addr)) {
                        returnstatus = TRUE;
                    }
                    AME_spinunlock(pAMEData);
			    }	   
            } else {               
                AME_spinlock(pAMEData);
                returnstatus = AME_Queue_pending_SCSI_Command(pAMEQueueData,RAID_ID,ReqType,Request_Buffer_Phy_addr);
                AME_spinunlock(pAMEData); 
            }
            break;
        
        default:
            returnstatus = FALSE;
            AME_Msg_Err("AME Queue Error:  unknow cmd type (0x%x)\n",ReqType);
            break;
    }
    
    return returnstatus;
}


AME_U32 AME_Queue_Command_SCSI_complete(pAME_Data pAMEData,AME_U32 RAID_ID,AME_U32 BufferID)
{
    AME_U32             returnstatus;
    AME_U32             issueCmd = FALSE;
    AME_Queue_Request   AMEQueueRequest;
    pAME_Queue_DATA pAMEQueueData = (pAME_Queue_DATA)pAMEData->pAMEQueueData;
    pAME_Queue_List pNode = NULL;
    
    if((pAMEData->Device_ID == AME_6101_DID_DAS) || (pAMEData->Device_ID == AME_6201_DID_DAS)){
       return FALSE; 
    }

    AME_spinlock(pAMEData);
    if (AME_Check_is_3108_new_type(pAMEData, RAID_ID) == TRUE) {
        //In new scsi type, V2, V2_R50, check Buffer ID
        returnstatus = AME_Queue_Get_Sended_SCSI_Command(pAMEQueueData,BufferID,RAID_ID); //Sam: performance turning.
    } else {
        returnstatus = AME_Queue_Get_Sended_SCSI_Command(pAMEQueueData,pAMEData->AMEBufferData[BufferID].Request_Buffer_Phy_addr,RAID_ID);
    }
    
	if (FALSE == returnstatus) {
        AME_Msg_Err("AME_Queue_Command_SCSI_complete:  Command not found in sending queue. BufferID = %d , Request_Buffer_Phy_addr = 0x%x , RAID_ID = %d\n",BufferID,pAMEData->AMEBufferData[BufferID].Request_Buffer_Phy_addr,RAID_ID);     	    
	}
	
	//R50 scsi cmd, count to Master Raid
	if (AME_Check_is_Raid_R50(pAMEData) == TRUE) {
	    if (pAMEQueueData->AME_Queue_Raid_CMD_Count[0] != 0) {
	        pAMEQueueData->AME_Queue_Raid_CMD_Count[0]--;
        } else {
            AME_Msg_Err("AME_Queue_Command_SCSI_complete:  AME_Queue_Raid_CMD_Count[%d] = 0, can't -1 \n",0);
        }
    } else {
	    if (pAMEQueueData->AME_Queue_Raid_CMD_Count[RAID_ID] != 0) {
	        pAMEQueueData->AME_Queue_Raid_CMD_Count[RAID_ID]--;
        } else {
            AME_Msg_Err("AME_Queue_Command_SCSI_complete:  AME_Queue_Raid_CMD_Count[%d] = 0, can't -1 \n",RAID_ID);
        }
    }

	returnstatus = AME_Queue_Get_Pended_SCSI_Command(pAMEQueueData,&AMEQueueRequest,&pNode);
	AME_spinunlock(pAMEData); 
    
    if (TRUE == returnstatus) {
		AME_U32 MAX_concurrent_command;
		AME_U32 ReqType;

        AME_spinlock(pAMEData);
        MAX_concurrent_command = 0;
        ReqType = AMEQueueRequest.ReqType;
        
        MAX_concurrent_command = AME_Queue_Get_Max_cmd(ReqType);
        if (MAX_concurrent_command == 0) {
            AME_Msg_Err("AME_Queue_Command_SCSI_complete:  Get unknow pended scsi cmd type (0x%x)\n",ReqType);
        }

		if((pAMEQueueData->AME_Queue_Raid_CMD_Count[AMEQueueRequest.Raid_ID] < MAX_concurrent_command) &&
		   (AME_NT_Raid_Bar_lost_Check(pAMEData,AMEQueueRequest.Raid_ID) == FALSE)) {
            issueCmd = TRUE;    
			pAMEQueueData->AME_Queue_Raid_CMD_Count[AMEQueueRequest.Raid_ID]++;
        	AME_Queue_add_tail(&pAMEQueueData->AME_Queue_List_SCSI_Sending,pNode);
        } else {   
            AME_Queue_add_head(&pAMEQueueData->AME_Queue_List_SCSI_Wait,pNode);
        }
        AME_spinunlock(pAMEData);

		if(TRUE == issueCmd) {
            
            if (TRUE == AME_Check_is_NT_Mode(pAMEData)) {
                returnstatus = AME_NT_write_Host_Inbound_Queue_Port_0x40(pAMEData,AMEQueueRequest.Raid_ID,AMEQueueRequest.Request_Buffer_Phy_addr);
            } else {
                returnstatus = AME_write_Host_Inbound_Queue_Port_0x40(pAMEData,(AME_U32)AMEQueueRequest.Request_Buffer_Phy_addr);
            }
            
            AME_spinlock(pAMEData);       
            if (FALSE == returnstatus) {
                // lost one cmd , but why can't send cmd to Raid ...
                AME_Msg_Err("AME_Queue_Command_SCSI_complete:  send next cmd fail\n");
                pAMEQueueData->AME_Queue_Raid_CMD_Count[AMEQueueRequest.Raid_ID]--;
                AME_Queue_del_entry(pNode);
                AME_Queue_add_head(&pAMEQueueData->AME_Queue_List_SCSI_Wait,pNode);
            }
            AME_spinunlock(pAMEData);
        } 
    }
    
    return TRUE;
}


AME_U32 AME_Queue_Command_InBand_complete(pAME_Data pAMEData)
{
    AME_U32             returnstatus;
    AME_Queue_Request   AMEQueueRequest;
    pAME_Queue_DATA pAMEQueueData = (pAME_Queue_DATA)pAMEData->pAMEQueueData;
    
    AME_spinlock(pAMEData);
	returnstatus = AME_Queue_Get_Pended_InBand_Command(pAMEQueueData,&AMEQueueRequest);
	if (FALSE == returnstatus) {
	    pAMEQueueData->AME_Queue_InBand_Count--;
	}
    AME_spinunlock(pAMEData);
    
    if (TRUE == returnstatus) {
        
        pAMEQueueData->Current_InBand_InBand_callback = AMEQueueRequest.callback;
        
        if ((AMEQueueRequest.ReqType == ReqType_NT_InBand_internal) || (AMEQueueRequest.ReqType == ReqType_NT_InBand_external)) {
            returnstatus = AME_NT_write_Host_Inbound_Queue_Port_0x40(pAMEData,AMEQueueRequest.Raid_ID,AMEQueueRequest.Request_Buffer_Phy_addr);
        } else {
            returnstatus = AME_write_Host_Inbound_Queue_Port_0x40(pAMEData,(AME_U32)AMEQueueRequest.Request_Buffer_Phy_addr);
        }
        
        if (FALSE == returnstatus) {
            // lost one cmd , but why can't send cmd to Raid ...
            AME_Msg_Err("AME_Queue_Command_InBand_complete:  send next cmd fail\n");
            AME_spinlock(pAMEData);
            pAMEQueueData->AME_Queue_InBand_Count--;
            AME_spinunlock(pAMEData);
        }
    }
    
    return TRUE;
}


AME_U32 AME_Queue_Abort_InBand_Command(pAME_Queue_DATA pAMEQueueData,AME_U64 Request_Buffer_Phy_addr)
{
    pAME_Queue_Buffer   pAMEQueueBuffer;
    pAME_Queue_List pHead = &pAMEQueueData->AME_Queue_List_InBand;
    pAME_Queue_List pNext = pHead;
    
    // AME_Check_is_3108_new_type = TRUE, Request_Buffer_Phy_addr not address, used other check mathod.
    pAME_Data pAMEData  = (pAME_Data)pAMEQueueData->pAMEData;
    AME_U32 i,Abort_BufferID=0,Abort_RaidID=0;
    
    for (i=1; i<pAMEData->Total_Support_command; i++) {
        if (Request_Buffer_Phy_addr == pAMEData->AMEBufferData[i].Request_Buffer_Phy_addr) {
            Abort_BufferID = i;
            Abort_RaidID = pAMEData->AMEBufferData[i].Raid_ID;
            break;
        }
        if (i == (pAMEData->Total_Support_command - 1)) {
            return FALSE;
        }
    }
    
    // check is current InBand cmd first
    if (pAMEQueueData->Current_InBand_Request_Buffer_Phy_addr == Request_Buffer_Phy_addr) {
        pAMEQueueData->Current_InBand_Is_Abort = TRUE;
        return TRUE;
    }
    
    while (1)
    {
        pNext = (pAME_Queue_List)pNext->pNext;
        
        if (pNext == pHead) {
            return FALSE;
        }
        
        pAMEQueueBuffer = (pAME_Queue_Buffer)pNext;
        if (pAMEQueueBuffer->Request_Buffer_Phy_addr == Request_Buffer_Phy_addr) {
            
            AME_Queue_del_entry(pNext);
            
            // Clear QueueBuffer
            pAMEQueueBuffer->ReqInUse = FALSE;
            pAMEQueueBuffer->Raid_ID  = 0;
            pAMEQueueBuffer->ReqType  = 0;
            pAMEQueueBuffer->Request_Buffer_Phy_addr = 0;
            
            AME_Queue_add_tail(&pAMEQueueData->AME_Queue_List_Free,pNext);
            return TRUE;
        }
        
        // AME_Check_is_3108_new_type Request_Buffer_Phy_addr not address, used other check mathod.
        if (AME_Check_is_3108_new_type(pAMEData, Abort_RaidID) == TRUE)
        {
            AME_U32 BufferID;
            
            // SgCount            :  8;     bits  0:7
            // PCISerialNumber    :  5;     bits  8:12
            // Reserved1          :  3;     bits  13:15
            // AddressIndex       : 16;     bits  16:31
            BufferID = ((AME_U32)pAMEQueueBuffer->Request_Buffer_Phy_addr&0xFFFF0000) >> 16;
            
            if (Abort_BufferID == BufferID) {
                AME_Queue_del_entry(pNext);
            
                // Clear QueueBuffer
                pAMEQueueBuffer->ReqInUse = FALSE;
                pAMEQueueBuffer->Raid_ID  = 0;
                pAMEQueueBuffer->ReqType  = 0;
                pAMEQueueBuffer->Request_Buffer_Phy_addr = 0;
                
                AME_Queue_add_tail(&pAMEQueueData->AME_Queue_List_Free,pNext);
                return TRUE;
            }
        }
        
    }
    
}


AME_U32 AME_Queue_Abort_SCSI_Command(pAME_Queue_DATA pAMEQueueData,AME_U64 Request_Buffer_Phy_addr)
{
    pAME_Queue_Buffer   pAMEQueueBuffer;
    pAME_Queue_List pHead = &pAMEQueueData->AME_Queue_List_SCSI_Wait;
    pAME_Queue_List pNext = pHead;
    AME_U32 ret = FALSE;
    
    // AME_Check_is_3108_new_type = TRUE, Request_Buffer_Phy_addr not address, used other check mathod.
    pAME_Data pAMEData  = (pAME_Data)pAMEQueueData->pAMEData;
    AME_U32 i,Abort_BufferID=0,Abort_RaidID=0;
    
    for (i=1; i<pAMEData->Total_Support_command; i++) {
        if (Request_Buffer_Phy_addr == pAMEData->AMEBufferData[i].Request_Buffer_Phy_addr) {
            Abort_BufferID = i;
            Abort_RaidID = pAMEData->AMEBufferData[i].Raid_ID;
            break;
        }
        if (i == (pAMEData->Total_Support_command - 1)) {
            return FALSE;
        }
    }
    
    while (1)
    {
        pNext = (pAME_Queue_List)pNext->pNext;
        
        if (pNext == pHead) {
            break;
        }
        
        pAMEQueueBuffer = (pAME_Queue_Buffer)pNext;
        if (pAMEQueueBuffer->Request_Buffer_Phy_addr == Request_Buffer_Phy_addr) {
            AME_Queue_del_entry(pNext);
            
            // Clear QueueBuffer
            pAMEQueueBuffer->ReqInUse = FALSE;
            pAMEQueueBuffer->Raid_ID  = 0;
            pAMEQueueBuffer->ReqType  = 0;
            pAMEQueueBuffer->Request_Buffer_Phy_addr = 0;
            
            AME_Queue_add_tail(&pAMEQueueData->AME_Queue_List_Free,pNext);
            return TRUE;
        }
        
        // AME_Check_is_3108_new_type Request_Buffer_Phy_addr not address, used other check mathod.
        if (AME_Check_is_3108_new_type(pAMEData, Abort_RaidID) == TRUE)
        {
            AME_U32 BufferID;
            
            // SgCount            :  8;     bits  0:7
            // PCISerialNumber    :  5;     bits  8:12
            // Reserved1          :  3;     bits  13:15
            // AddressIndex       : 16;     bits  16:31
            BufferID = ((AME_U32)pAMEQueueBuffer->Request_Buffer_Phy_addr&0xFFFF0000) >> 16;
            
            if (Abort_BufferID == BufferID) {
                AME_Queue_del_entry(pNext);
            
                // Clear QueueBuffer
                pAMEQueueBuffer->ReqInUse = FALSE;
                pAMEQueueBuffer->Raid_ID  = 0;
                pAMEQueueBuffer->ReqType  = 0;
                pAMEQueueBuffer->Request_Buffer_Phy_addr = 0;
                
                AME_Queue_add_tail(&pAMEQueueData->AME_Queue_List_Free,pNext);
                return TRUE;
            }
        }
    }
    
    
    pHead = &pAMEQueueData->AME_Queue_List_SCSI_Sending;
    pNext = pHead;
    
    while (1)
    {
        pNext = (pAME_Queue_List)pNext->pNext;
        
        if (pNext == pHead) {
            break;
        }
        
        pAMEQueueBuffer = (pAME_Queue_Buffer)pNext;
        if (pAMEQueueBuffer->Request_Buffer_Phy_addr == Request_Buffer_Phy_addr) {
            AME_Queue_del_entry(pNext);
            
            // Clear QueueBuffer
            pAMEQueueBuffer->ReqInUse = FALSE;
            pAMEQueueBuffer->Raid_ID  = 0;
            pAMEQueueBuffer->ReqType  = 0;
            pAMEQueueBuffer->Request_Buffer_Phy_addr = 0;
            
            AME_Queue_add_tail(&pAMEQueueData->AME_Queue_List_Free,pNext);
            return TRUE;
        }
        
        // AME_Check_is_3108_new_type Request_Buffer_Phy_addr not address, used other check mathod.
        if (AME_Check_is_3108_new_type(pAMEData, Abort_RaidID) == TRUE)
        {
            AME_U32 BufferID;
            
            // SgCount            :  8;     bits  0:7
            // PCISerialNumber    :  5;     bits  8:12
            // Reserved1          :  3;     bits  13:15
            // AddressIndex       : 16;     bits  16:31
            BufferID = ((AME_U32)pAMEQueueBuffer->Request_Buffer_Phy_addr&0xFFFF0000) >> 16;
            
            if (Abort_BufferID == BufferID) {
                AME_Queue_del_entry(pNext);
            
                // Clear QueueBuffer
                pAMEQueueBuffer->ReqInUse = FALSE;
                pAMEQueueBuffer->Raid_ID  = 0;
                pAMEQueueBuffer->ReqType  = 0;
                pAMEQueueBuffer->Request_Buffer_Phy_addr = 0;
                
                AME_Queue_add_tail(&pAMEQueueData->AME_Queue_List_Free,pNext);
                return TRUE;
            }
        }
        
    }
    
    return FALSE;
}


AME_U32 AME_Queue_Abort_Command(pAME_Data pAMEData,AME_U64 Request_Buffer_Phy_addr)
{
    AME_U32   returnstatus;
    pAME_Queue_DATA pAMEQueueData = (pAME_Queue_DATA)pAMEData->pAMEQueueData;
    
    
    AME_spinlock(pAMEData);
    returnstatus = AME_Queue_Abort_SCSI_Command(pAMEQueueData,Request_Buffer_Phy_addr);
    AME_spinunlock(pAMEData);
    
    if (TRUE == returnstatus) {
        AME_Msg_Err("Abort request in SCSI list queue done ...\n");
        return TRUE;
    }
    
    AME_spinlock(pAMEData);
    returnstatus = AME_Queue_Abort_InBand_Command(pAMEQueueData,Request_Buffer_Phy_addr);
    AME_spinunlock(pAMEData);
    
    if (TRUE == returnstatus) {
        AME_Msg_Err("Abort request in InBand list queue done ...\n");
        return TRUE;
    }
    
    return FALSE;
}


AME_U32 AME_Queue_Send_Pended_SCSI_Command(pAME_Data pAMEData)
{
    AME_U32             returnstatus;
    AME_U32             issueCmd = FALSE;
    AME_Queue_Request   AMEQueueRequest;
    pAME_Queue_DATA pAMEQueueData = (pAME_Queue_DATA)pAMEData->pAMEQueueData;
    pAME_Queue_List pNode = NULL;
    
    if((pAMEData->Device_ID == AME_6101_DID_DAS) || (pAMEData->Device_ID == AME_6201_DID_DAS)){
       return FALSE; 
    }

    AME_spinlock(pAMEData);
	returnstatus = AME_Queue_Get_Pended_SCSI_Command(pAMEQueueData,&AMEQueueRequest,&pNode);   
	AME_spinunlock(pAMEData); 
    
    if (TRUE == returnstatus) {
		AME_U32 MAX_concurrent_command;
		AME_U32 ReqType;

        AME_spinlock(pAMEData);
        MAX_concurrent_command = 0;
        ReqType = AMEQueueRequest.ReqType;
        
        MAX_concurrent_command = AME_Queue_Get_Max_cmd(ReqType);
        if (MAX_concurrent_command == 0) {
            AME_Msg_Err("AME_Queue_Send_Pended_SCSI_Command:  Get unknow pended scsi cmd type (0x%x)\n",ReqType);
        }

		if((pAMEQueueData->AME_Queue_Raid_CMD_Count[AMEQueueRequest.Raid_ID] < MAX_concurrent_command) &&
		   (AME_NT_Raid_Bar_lost_Check(pAMEData,AMEQueueRequest.Raid_ID) == FALSE)) {
            issueCmd = TRUE;    
			pAMEQueueData->AME_Queue_Raid_CMD_Count[AMEQueueRequest.Raid_ID]++;
        	AME_Queue_add_tail(&pAMEQueueData->AME_Queue_List_SCSI_Sending,pNode);
        } else {   
            AME_Queue_add_head(&pAMEQueueData->AME_Queue_List_SCSI_Wait,pNode);
        }
        AME_spinunlock(pAMEData);

		if(TRUE == issueCmd) {
            
            if(TRUE == AME_Check_is_NT_Mode(pAMEData)) {                   
                returnstatus = AME_NT_write_Host_Inbound_Queue_Port_0x40(pAMEData,AMEQueueRequest.Raid_ID,AMEQueueRequest.Request_Buffer_Phy_addr);
            } else {
                returnstatus = AME_write_Host_Inbound_Queue_Port_0x40(pAMEData,(AME_U32)AMEQueueRequest.Request_Buffer_Phy_addr);
            }
            
            AME_spinlock(pAMEData);       
            if (FALSE == returnstatus) {
                // lost one cmd , but why can't send cmd to Raid ...
                AME_Msg_Err("AME_Queue_Command_SCSI_complete:  send next cmd fail\n");
                pAMEQueueData->AME_Queue_Raid_CMD_Count[AMEQueueRequest.Raid_ID]--;
                AME_Queue_del_entry(pNode);
                AME_Queue_add_head(&pAMEQueueData->AME_Queue_List_SCSI_Wait,pNode);
            }
            AME_spinunlock(pAMEData);
        } 
    }
    
    return TRUE;
}

